/****************************************/
/* Cool Herders                         */
/* A demented muse by Brendan Wiese     */
/* http://starbow.org                   */
/* Programmed by Mike Brent             */
/* http://tursi.yiffco.com              */
/* bmp.c                                */
/****************************************/

/* This BMP reader code is not portable to a big-endian system!    */
/* It does a lot of weird things specific to my code, and only     */
/* knows 24-bit uncompressed Windows BMPs. I may change that. ;)   */
/* It standardizes on 256x256 tile blocks. Anything larger is      */
/* loaded into multiple blocks. A 256x256 block takes 128k of VRAM */
/* The blocks are located one right after the other in memory, and */
/* are organized such that horizontal rows are completed before    */
/* the next vertical row. This is all a bit hacky, code will be    */
/* cleaned up when I get more time. :)                             */
/* code is loaded into ARGB1555 data.	                           */
/* Adapted from the much more flexible Allegro load_bmp routines   */
/* by Jonas Petersen                                               */

/* modified for use with non-os KOS-1.1.4+ - PBoese */

#include <kos.h>
#include "bmp.h"

#undef FILE
#define FILE int

int read_bmfileheader(FILE f, BITMAPFILEHEADER *fileheader);
int read_win_bminfoheader(FILE f, BITMAPINFOHEADER *infoheader);
void read_24bit_line(int length, FILE f, BITMAP *bmp, int line);
void read_image(FILE f, BITMAP *bmp, BITMAPINFOHEADER *infoheader);

/* read_bmfileheader:
*  Reads a BMP file header and check that it has the BMP magic number.
*/
int read_bmfileheader(FILE f, BITMAPFILEHEADER *fileheader)
{
	fs_read(f, &fileheader->bfType, 2);
	fs_read(f, &fileheader->bfSize, 4);
	fs_read(f, &fileheader->bfReserved1, 2);
	fs_read(f, &fileheader->bfReserved2, 2);
	fs_read(f, &fileheader->bfOffBits, 4);
	
	if (fileheader->bfType != 19778) {	// BM in ASCII
		printf("Not a BMP file. (%d)\n", fileheader->bfType);
		return -1;
	}

	return 0;
}

/* read_win_bminfoheader:
*  Reads information from a BMP file header.
*/
int read_win_bminfoheader(FILE f, BITMAPINFOHEADER *infoheader)
{
	WINBMPINFOHEADER win_infoheader;
	
	fs_read(f, &win_infoheader.biWidth, 4);
	fs_read(f, &win_infoheader.biHeight, 4);
	fs_read(f, &win_infoheader.biPlanes, 2);
	fs_read(f, &win_infoheader.biBitCount, 2);
	fs_read(f, &win_infoheader.biCompression, 4);
	fs_read(f, &win_infoheader.biSizeImage, 4);
	fs_read(f, &win_infoheader.biXPelsPerMeter, 4);
	fs_read(f, &win_infoheader.biYPelsPerMeter, 4);
	fs_read(f, &win_infoheader.biClrUsed, 4);
	fs_read(f, &win_infoheader.biClrImportant, 4);
	
	infoheader->biWidth = win_infoheader.biWidth;
	infoheader->biHeight = win_infoheader.biHeight;
	infoheader->biBitCount = win_infoheader.biBitCount;
	infoheader->biCompression = win_infoheader.biCompression;

	infoheader->adjustedWidth=(((infoheader->biWidth-1)/256)+1)*256;
	infoheader->adjustedHeight=(((infoheader->biHeight-1)/256)+1)*256;
	infoheader->nXBlocks=infoheader->adjustedWidth/256;
	infoheader->nYBlocks=infoheader->adjustedHeight/256;

	if (infoheader->biBitCount!=24) {
		printf("Bitmap depth %d not supported. Only 24 bit is okay.\n", infoheader->biBitCount);
		return -1;
	}

	if (infoheader->biCompression!=0) {
		printf("Bitmap compression is not supported.\n");
		return -1;
	}

	printf("Bitmap loading to %d x %d, %d x %d blocks.\n", infoheader->adjustedWidth, infoheader->adjustedHeight, infoheader->nXBlocks, infoheader->nYBlocks);

	return 0;
}

/* read_24bit_line:
*  Support function for reading the 24 bit bitmap file format
*/
void read_24bit_line(int length, FILE f, BITMAP *bmp, int line)
{
	int i, nbytes;
	unsigned char r, g, b;
	unsigned short color;
	unsigned long address;
	char buf[8];
	
	nbytes=0;
	
	for (i=0; i<length; i++) {
		fs_read(f, &b, 1);
		fs_read(f, &g, 1);
		fs_read(f, &r, 1);
		
		if ((r==0)&&(g==0)&&(b==0)) {
			color=0x0000;
		} else {
			color=((r>>3)<<10) | ((g>>3)<<5) | (b>>3) | 0x8000;
		}

		// Figure out which block and where to put this pixel
		// Base address of this block row
		address=(line/256)*((BLOCKSIZE/2) * (length/256+1));
		// Add base address of this block
		address+=((BLOCKSIZE/2)*(i/256));
		// Get line inside block
		address+=(line%256)*256;
		// Get offset on line
		address+=i%256;
		// that should do it.
		bmp->data[address]=color;
	}

	nbytes = nbytes % 4;
	if (nbytes != 0) {
		nbytes=4-nbytes;
		fs_read(f, buf, nbytes);
	}
}

/* read_image:
*  For reading the noncompressed BMP image format.
*/
void read_image(FILE f, BITMAP *bmp, BITMAPINFOHEADER *infoheader)
{
	int i, line;
	
	for (i=0; i<(int)infoheader->biHeight; i++) {
		line = i;
		read_24bit_line(infoheader->biWidth, f, bmp, infoheader->biHeight-i-1);
	}
}

/* load_bmp:
*  Loads a Windows 24-BMP file, returning a 16-bit bitmap structure
*/
BITMAP *load_bmp(char *filename)
{
	BITMAPFILEHEADER fileheader;
	BITMAPINFOHEADER infoheader;
	FILE f;
	BITMAP *bmp;
	int ncol;
	unsigned long biSize;
	int bpp, dest_depth;
	int line, i;

	f = fs_open(filename, O_RDONLY);
	if (!f) {
		printf("Failed to open %s\n", filename);
		return NULL;
	} else {
		printf("Opened %s\n", filename);
	}

	if (read_bmfileheader(f, &fileheader) != 0) {
		fs_close(f);
		return NULL;
	}

	fs_read(f, (void*)&biSize, 4);

	if (biSize == WININFOHEADERSIZE) {
		if (read_win_bminfoheader(f, &infoheader) != 0) {
			fs_close(f);
			return NULL;
		}
	}
	else 
	{
		printf("Bitmap header size %d unknown. Only Windows files supported.\n", biSize);
		fs_close(f);
		return NULL;
	}

	printf("Image: %d x %d\n", infoheader.biWidth, infoheader.biHeight);
	bmp = (BITMAP*)malloc(sizeof(BITMAP) + (infoheader.nXBlocks * infoheader.nYBlocks * BLOCKSIZE));
	if (!bmp) {
		printf("Malloc failed\n");
		fs_close(f);
		return NULL;
	}
	printf("..malloced..\n");
	memset(bmp, 0, sizeof(BITMAP) + (infoheader.nXBlocks * infoheader.nYBlocks * BLOCKSIZE));

	bmp->nWidth=infoheader.biWidth;
	bmp->nHeight=infoheader.biHeight;
	bmp->adjustedWidth=infoheader.adjustedWidth;
	bmp->adjustedHeight=infoheader.adjustedHeight;
	bmp->nXBlocks=infoheader.nXBlocks;
	bmp->nYBlocks=infoheader.nYBlocks;

	read_image(f, bmp, &infoheader);

	fs_close(f);

	return bmp;
}

